"""
API endpoint agent.
"""
import argparse
import os
import logging

from twisted.internet.defer import inlineCallbacks, returnValue

from juju.agents.base import BaseAgent
from juju.environment.config import EnvironmentsConfig
from juju.errors import JujuError
from juju.lib.websockets import WebSocketsResource
from juju.rapi.transport.ws import WebSocketAPIFactory
from juju.state.auth import make_ace

from twisted.internet import ssl
from twisted.web.static import Data
from twisted.web.server import Site
from zookeeper import NoNodeException

log = logging.getLogger("juju.agents.api")
KEYS_DEFAULT_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'keys'))


def keys_directory(argument):
    """Validate the path argument and return *(private key, certificate)*."""
    path = os.path.abspath(os.path.expanduser(argument))
    if os.path.isdir(path):
        key = os.path.join(path, 'juju.key')
        cert = os.path.join(path, 'juju.crt')
        if os.path.isfile(key) and os.path.isfile(cert):
            # The order is significant here. These paths will be passed to
            # ``twisted.internet.ssl.DefaultOpenSSLContextFactory``.
            return key, cert
        msg = 'cannot find the juju.key and juju.crt files in %r' % argument
        raise argparse.ArgumentTypeError(msg)
    msg = '%r: no such directory' % argument
    raise argparse.ArgumentTypeError(msg)


class APIEndpointAgent(BaseAgent):

    name = "juju-api"

    @inlineCallbacks
    def start(self):
        log.info("Starting api server")

        self.environment = yield self.configure_environment()
        self.provider = yield self._get_provider()
        log.debug("Received environment data.")
        root = Data("Juju API Server\n", "text/plain")
        self.ws_factory = WebSocketAPIFactory(
            self.config['zookeeper_servers'], self.provider)
        yield self.ws_factory.startFactory()

        root.putChild('ws', WebSocketsResource(self.ws_factory))

        from twisted.internet import reactor

        port = int(self.config.get('port', 80))
        site = Site(root)
        if self.config.get('secure'):
            keys = self.config['keys'] or keys_directory(KEYS_DEFAULT_PATH)
            context_factory = ssl.DefaultOpenSSLContextFactory(*keys)
            self.api_socket = reactor.listenSSL(port, site, context_factory)
            log.info("Secure websocket active")
        else:
            self.api_socket = reactor.listenTCP(port, site)
            log.info("Websocket active")

    def stop(self):
        if self.api_socket:
            self.ws_factory.stopFactory()
            self.api_socket.stopListening()
            self.api_socket = None

    def _get_provider(self):
        return self.environment.get_machine_provider()

    # TODO Refactory into common for provisioning agent
    @inlineCallbacks
    def configure_environment(self):
        """The provisioning agent configure its environment on start or change.

        The environment contains the configuration th agent needs to interact
        with its machine provider, in order to do its work. This configuration
        data is deployed lazily over an encrypted connection upon first usage.

        The agent waits for this data to exist before completing its startup.
        """
        try:
            get_d, watch_d = self.client.get_and_watch("/environment")
            environment_data, stat = yield get_d
        except NoNodeException:
            # Wait till the environment node appears. play twisted gymnastics
            exists_d, watch_d = self.client.exists_and_watch("/environment")
            stat = yield exists_d
            if stat:
                environment = yield self.configure_environment()
            else:
                watch_d.addCallback(
                    lambda result: self.configure_environment())
            if not stat:
                environment = yield watch_d
            returnValue(environment)

        # Lazy initialize login nodes if not present.
        yield self._initialize_login()

        config = EnvironmentsConfig()
        config.parse(environment_data)
        returnValue(config.get_default())

    @inlineCallbacks
    def _initialize_login(self):
        children = yield self.client.get_children("/")
        if "login" in children:
            return

        acls, stat = yield self.client.get_acl("/initialized")
        admin_acl = None

        for a in acls:
            if a['id'].startswith('admin'):
                admin_acl = a
                break

        yield self.client.create(
            "/login", acls=[make_ace(admin_acl['id'], all=True)])

    def configure(self, options):
        super(APIEndpointAgent, self).configure(options)
        if not options.get("port"):
            msg = ("--port must be provided in the command line")
            raise JujuError(msg)
        if options.get("keys") and not options.get("secure"):
            msg = ("in addition to providing keys, a secure connection must "
                   "be explicitly enabled using --secure")
            raise JujuError(msg)
        self.api_socket = None

    @classmethod
    def setup_options(cls, parser):
        super(APIEndpointAgent, cls).setup_options(parser)
        port = os.environ.get("JUJU_API_PORT", "")
        default_port = port.isdigit() and int(port) or 80
        parser.add_argument(
            "--port", default=default_port, type=int)
        parser.add_argument("--secure", action='store_true')
        parser.add_argument("--keys", default=os.getenv('JUJU_API_KEYS'),
            type=keys_directory, metavar='PATH')
        return parser

if __name__ == '__main__':
    APIEndpointAgent.run()
