"""
Annotation management
---------------------

Design choices
==============

 In the interest of simplicity and facility given strategic goals for
 pyjuju, the implementation here is kept as simple as possible with a
 target annotation count (items and contexts annotated) under 5k.

 The primary purpose is to support the rapid rollout in support of
 integration efforts for juju gui and landcape.


API Usage
=========

Add an annotation to an object::

  >> annotations.set('mysql/0', {'machine_id': ''})

Annotations are cumulative::

  >> annotations.set('mysql/0', '{'restart': True})
  >> annotations.set('mysql/0', {'gui.x': 0, 'gui.y': 1})

Annotations can be deleted::

  >> annotations.remove('mysql/1', keys=['restart'])
  >> annotations.remove('mysql')


Annotations can be retrieved against a context::

  >> data = annotations.get('mysql/0')
  >> print data

  {'restart': True', 'machine_id': 21, 'x': 0, 'y': 21}


Annotations changes can be retrieved in a delta::

  >> stream = annotations.stream()
  >> stream.next()
  >> annotations.set('mysql/0', {'a': 21})
  >> stream.next()
  {'mysql/0', {'a': 21}

Delta returns current value for any context since last checked. For
contexts removed, that will be an empty dictionary.

Annotations are expected to be high rate of change. As a result the
delta api here is explicit instead of watch callback, especially
given its driver is timer based. This avoids much spurious change
processing.

Annotation are disjoint from the items under management, so they need
periodic garbage collection to remove data for context no longer extant.

  >> annotations.gc()

Care is taken to never return stale data for an item even without
gc. The gc is solely for space management.
"""
import copy
import json
import logging
import zlib
import zookeeper

from txzookeeper.utils import retry_change
from twisted.internet.defer import inlineCallbacks, returnValue

from juju.state.base import StateBase
from juju.state.errors import InvalidAnnotationContext, StateError
from juju.state.topology import InternalTopologyError


class AnnotationManager(StateBase):

    ANNOTATION_PATH = "/annotations"

    def _dump(self, data):
        """Serialize the given data structure to compressed json string."""
        return zlib.compress(json.dumps(data))

    def _load(self, content):
        """Load the data from a compressed json serialized content."""
        if not content:
            return {}
        return json.loads(zlib.decompress(content))

    @inlineCallbacks
    def _verify_context(self, context):
        """Verify the given context is valid in the current topology.

        Raises a InvalidAnnotationContext if the context is not valid.
        Returns an annotation key for the context if its valid.
        """
        topology = yield self._read_topology()
        key = self._get_key(context, topology)
        if key is None:
            raise InvalidAnnotationContext(context)
        returnValue(key)

    def _get_key(self, context, topology):
        """Return a key for the context if its valid in the given topology.

        If not return None.
        """
        if context == "env":
            return "env"

        if isinstance(context, int) or context.isdigit():
            internal_id = "machine-%010d" % int(context)
            if topology.has_machine(internal_id):
                return internal_id

        service_id = topology.find_service_with_name(context)
        if service_id is not None:
            return service_id

        if "/" in context:
            try:
                uid = topology.get_service_unit_id_from_name(context)
            except InternalTopologyError:
                pass
            else:
                if uid:
                    return uid

    def _get_context(self, key, topology):
        """Retrieve the context name for a given key.
        """
        if key.startswith('machine'):
            return int(key.split('-')[-1])
        elif key.startswith('unit'):
            try:
                return topology.get_service_unit_name_from_id(key)
            except InternalTopologyError:
                return None
        elif key.startswith('service') and topology.has_service(key):
            return topology.get_service_name(key)
        elif key == "env":
            return key

    @inlineCallbacks
    def set(self, context, data):
        """Set some annotation data on a context.

        Annotations set on a context are cumulative/merged with
        existing annotations.
        """
        if not isinstance(data, dict):
            raise StateError("Invalid annotation data for %s: -> %r" % (
                context, data))

        if filter(None, [not isinstance(k, basestring) for k in data.keys()]):
            raise StateError(
                "Invalid annotation keys must be strings -> %r", data)

        key = yield self._verify_context(context)

        def _update(content, stat):
            store = self._load(content)
            ctx_store = store.setdefault(key, {})
            ctx_store.update(data)
            return self._dump(store)

        yield retry_change(self._client, self.ANNOTATION_PATH, _update)
        returnValue(True)

    @inlineCallbacks
    def get(self, context=None):
        """Retrieve annotations for a given context, or all annotations.
        """
        k = context and (yield self._verify_context(context))
        try:
            content, stat = yield self._client.get(self.ANNOTATION_PATH)
        except zookeeper.NoNodeException:
            returnValue({})
        store = self._load(content)

        if context:
            returnValue(store.get(k, {}))

        result = {}
        topology = yield self._read_topology()
        for k, v in store.items():
            c = self._get_context(k, topology)
            if c is None:
                continue
            result[c] = v
        returnValue(result)

    @inlineCallbacks
    def remove(self, context, keys=None):
        ckey = yield self._verify_context(context)

        def _update(content, stat):
            data = self._load(content)
            if not ckey in data:
                return content
            if not keys:
                del data[ckey]
                return self._dump(data)
            for k in keys:
                data[ckey].pop(k, None)
            return self._dump(data)

        yield retry_change(self._client, self.ANNOTATION_PATH, _update)
        returnValue(True)

    @inlineCallbacks
    def gc(self):

        @inlineCallbacks
        def _update(content, stat):
            valid_keys = []
            topology = yield self._read_topology()
            services = topology.get_services()
            valid_keys.extend(services)

            for k in services:
                valid_keys.extend(topology.get_service_units(k))

            valid_keys.extend(topology.get_machines())

            store = self._load(content)
            used_keys = store.keys()
            waste = set(used_keys) - set(valid_keys)
            for k in waste:
                store.pop(k, None)
            returnValue(self._dump(store))

        yield retry_change(self._client, self.ANNOTATION_PATH, _update)

    def stream(self):
        return AnnotationStream(self._client, self)


class AnnotationStream(object):

    def __init__(self, client, annotations):
        self._client = client
        self._annotations = annotations
        self._last = None
        self._log = logging.getLogger("annotation.stream")

    @inlineCallbacks
    def next(self):
        if not self._last:
            self._last = yield self._annotations.get()
            returnValue(copy.deepcopy(self._last))

        delta = {}
        data = yield self._annotations.get()
        keys = set(data.keys())
        last_keys = set(self._last.keys())

        new = keys - last_keys
        for n in new:
            delta[n] = data[n]

        # TODO - requires some surgery to rapi/delta.py, for now
        # only update those extant with removed data.

        # here's a scenario of concern for the todo, with s == service
        # s-a added, annotated, s-a removed, s-a added.. in the span
        # of a delta push.... correct would be no annotations for
        # s-a. the annotations manager will properly recognize two
        # instances, and omit for second instance of s-a.  status
        # consumers can't distinguish since they have the same user
        # label/name. we must then delete or update delta stream to
        # distinguish else we get ghost annotations.
        t = yield self._annotations._read_topology()
        removed = last_keys - keys
        for r in removed:
            if self._annotations._get_key(r, t):
                delta[r] = {}

        existing = keys.intersection(last_keys)
        for k in existing:
            if data[k] != self._last[k]:
                delta[k] = data[k]

        self._last = data
        returnValue(delta)
