from twisted.internet.defer import inlineCallbacks

from juju.state.annotation import AnnotationManager
from juju.state.errors import InvalidAnnotationContext
from juju.state.topology import InternalTopology
from juju.state.tests.common import StateTestBase


class AnnotationManagerTest(StateTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(AnnotationManagerTest, self).setUp()
        self.topology = InternalTopology()
        self.annotations = AnnotationManager(self.client)

    @inlineCallbacks
    def test_get_set_annotation(self):
        self.topology.add_machine("machine-%010d" % 0)
        self.topology.add_service("service-1", "rethinkdb")
        self.topology.add_service_unit("service-1", "unit-1")
        yield self.set_topology(self.topology)

        yield self.annotations.set(0, {'series': 'precise'})
        yield self.annotations.set("rethinkdb", {'link': 'ref:21'})
        yield self.annotations.set("rethinkdb/0", {"usn": 42})

        # Can't set or get annotations against an invalid context.
        yield self.assertFailure(
            self.annotations.set("m-2", {'food': 'oatmeal'}),
            InvalidAnnotationContext)
        yield self.assertFailure(
            self.annotations.get("m-2"), InvalidAnnotationContext)
        yield self.assertFailure(
            self.annotations.set("rethinkdb/2", {'margin': 'error'}),
            InvalidAnnotationContext)
        yield self.assertFailure(
            self.annotations.get("rethinkdb/2"), InvalidAnnotationContext)
        yield self.assertFailure(
            self.annotations.get("mysql/2"), InvalidAnnotationContext)

        self.assertEqual(
            (yield self.annotations.get()),
            {0: {'series': 'precise'}, 'rethinkdb': {'link': 'ref:21'},
             "rethinkdb/0": {"usn": 42}})
        self.assertEqual(
            (yield self.annotations.get('rethinkdb')),
            {'link': 'ref:21'})
        self.assertEqual(
            (yield self.annotations.get('rethinkdb/0')),
            {'usn': 42})

    @inlineCallbacks
    def test_gc(self):
        self.topology.add_service("service-1", "rethinkdb")
        self.topology.add_service_unit("service-1", "unit-1")

        yield self.set_topology(self.topology)
        yield self.annotations.set("rethinkdb", {"greeting": "hello"})
        yield self.annotations.set("rethinkdb/0", {"greeting": "salu"})
        data = yield self.annotations.get()
        self.assertEqual(data, {"rethinkdb": {"greeting": "hello"},
                                "rethinkdb/0": {"greeting": "salu"}})

        # Kill the service
        self.topology.remove_service("service-1")
        yield self.set_topology(self.topology)

        # The stale annotation is immediately non visible
        data = yield self.annotations.get()
        self.assertFalse(data)

        # However the data is still extant in storage till its gc'd
        content, _ = yield self.client.get(self.annotations.ANNOTATION_PATH)
        data = self.annotations._load(content)
        self.assertTrue(data)

        yield self.annotations.gc()

        content, _ = yield self.client.get(self.annotations.ANNOTATION_PATH)
        data = self.annotations._load(content)
        self.assertFalse(data)

    @inlineCallbacks
    def test_remove(self):
        self.topology.add_service("service-1", "rethinkdb")
        self.topology.add_service_unit("service-1", "unit-1")

        yield self.set_topology(self.topology)
        yield self.annotations.set("rethinkdb/0", {
            "a": 1, "b": 2, "c": 3, "d": 4})

        yield self.annotations.remove("rethinkdb/0", ["a", "d"])
        data = yield self.annotations.get("rethinkdb/0")
        self.assertEqual(data, {"b": 2, "c": 3})

    @inlineCallbacks
    def test_stream(self):
        stream = self.annotations.stream()
        delta = yield stream.next()
        self.assertEqual(delta, {})

        self.topology.add_service("service-1", "rethinkdb")
        self.topology.add_service_unit("service-1", "unit-1")
        yield self.set_topology(self.topology)

        yield self.annotations.set("rethinkdb", {"slink": "example.com"})
        delta = yield stream.next()
        self.assertEqual(delta, {"rethinkdb": {"slink": "example.com"}})

        yield self.annotations.set(
            "rethinkdb/0", {"a": 1, "b": 2, "c": 3})
        yield self.annotations.remove("rethinkdb/0", ["a", "c"])
        delta = yield stream.next()
        self.assertEqual(delta, {"rethinkdb/0": {"b": 2}})

        yield self.annotations.remove("rethinkdb")
        delta = yield stream.next()
        self.assertEqual(delta, {"rethinkdb": {}})
